#include <locale.h>
#include "dfbiterm.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#ifdef HAVE_SELECT
#include <errno.h>
#include <iterm/unix/ttyio.h>
#endif

#include <directfb.h>
#include <pthread.h>

Iterm *pIterm;

#define DFBCHECK(x...)                                         \
{                                                            \
    DFBResult err = x;                                         \
                                                               \
    if (err != DFB_OK)                                         \
{                                                        \
        fprintf( stderr, "%s <%d>:\n\t", __FILE__, __LINE__ ); \
        DirectFBErrorFatal( #x, err );                         \
}                                                        \
}
static void *update_thread(void *);

void
dfb_end(void)
{
	pIterm->buffer->Release(pIterm->buffer);
	pIterm->font->Release(pIterm->font);
	pIterm->window->Release(pIterm->window);
	pIterm->layer->Release(pIterm->layer);
	pIterm->surface->Release(pIterm->surface);
	pIterm->dfb->Release(pIterm->dfb);
}

static void
exitDFbiterm(int sig)
{
  /* set new keyboard control parameters for new terminal */
  tcsetattr(0, TCSAFLUSH, &(pIterm->ntio));
  dfb_end();
  exit (0);
}

static void
dfbiterm_usage (void)
{
  fprintf (stdout, "Usage: iterm [ -a <fontfile> ] [ -H <height> ] [ - W <width> ] [ -n ] [ -v ]\n\
\n\
options:\n\
  -a <fontfile>\tascii text font\n\
  -H <height>\theight of the character cell\n\
  -W <width>\twidth of the character cell\n\
  -v\t\tprint version information and exit\n\
  -n\t\tno character map\n\
  -h\t\tthis help message\n");

  exit (0);
}

int
main (int argc, char *argv[])
{
	unsigned char buf[BUFSIZ + 1], *ascfontname, *mbfontname;
	int i;
	DFBFontDescription font_dsc;
	DFBSurfaceDescription dsc;
	DFBSurfacePixelFormat format;
	DFBWindowDescription desc;
	int nocharmap = 0;
	pthread_t thread;

	setlocale (LC_ALL, "");


	/* allocate memory for Iterm structure */
	pIterm = (Iterm *) calloc (1, sizeof (Iterm));
	if (!pIterm) {
		fprintf (stderr, "There is not enough memory available now.\n");
		exit (1);
	}
	ascfontname = DefaultFont;

	font_dsc.height = 16;
	font_dsc.width = 8;

	while ((i = getopt (argc, argv, "a:H:W:nhv")) != EOF)
		switch (i) {
		case 'a':
			if (optarg != NULL)
				ascfontname = optarg;
			else
				ascfontname = DefaultFont;
			break;
		case 'H':
			if (optarg) {
				font_dsc.height = atoi(optarg);
				if (!font_dsc.height) {
					font_dsc.height = 16;
				}
			}
			break;
		case 'W':
			if (optarg) {
				font_dsc.width = atoi(optarg);
				if (!font_dsc.width) {
					font_dsc.width = 8;
				}
			}
			break;
		case 'v':
			fprintf (stdout, "fbiterm %s\n", DFBITERM_VERSION);
			exit (0);
			break;
		case 'n':
			nocharmap = 1;
			break;
		case 'h':
		default:
			dfbiterm_usage ();
			break;
      		}
	if (optind < argc)
		dfbiterm_usage ();

	DFBCHECK (DirectFBInit (&argc, &argv));
	DFBCHECK (DirectFBCreate (&pIterm->dfb));
//	DFBCHECK (pIterm->dfb->SetCooperativeLevel (pIterm->dfb, DFSCL_FULLSCREEN));
	dsc.flags = DSDESC_CAPS;
	dsc.caps  = DSCAPS_PRIMARY;
	font_dsc.flags = DFDESC_HEIGHT | DFDESC_WIDTH;
	if (nocharmap) {
		font_dsc.flags |= DFDESC_ATTRIBUTES;
		font_dsc.attributes = DFFA_NOCHARMAP;
	}
	DFBCHECK (pIterm->dfb->CreateSurface(pIterm->dfb, &dsc, &pIterm->surface ));
	DFBCHECK (pIterm->surface->Clear(pIterm->surface, 0, 0, 0, 0xff));
	DFBCHECK (pIterm->surface->GetSize (pIterm->surface, &pIterm->screen_width, &pIterm->screen_height));
	DFBCHECK (pIterm->dfb->CreateFont (pIterm->dfb, ascfontname, &font_dsc, &pIterm->font));
	DFBCHECK (pIterm->surface->SetFont(pIterm->surface, pIterm->font));
	DFBCHECK (pIterm->font->GetHeight(pIterm->font, &pIterm->cell_height));

	pIterm->cell_width = font_dsc.width;
	pIterm->cols = pIterm->screen_width / pIterm->cell_width;
	pIterm->rows = pIterm->screen_height / pIterm->cell_height;

	DFBCHECK (pIterm->dfb->GetDisplayLayer(pIterm->dfb, DLID_PRIMARY, &pIterm->layer));
	DFBCHECK (pIterm->surface->GetPixelFormat(pIterm->surface, &format));

	if (format == DSPF_LUT8 || format == DSPF_ALUT44) {
		pIterm->indexed = 1;
		DFBCHECK (pIterm->surface->GetPalette(pIterm->surface, &pIterm->palette));
	}

	desc.width = pIterm->screen_width;
	desc.height = pIterm->screen_height;
	desc.posx = desc.posy = 0;
	desc.pixelformat = format;

	desc.flags =  DWDESC_WIDTH | DWDESC_HEIGHT | DWDESC_PIXELFORMAT | DWDESC_POSX | DWDESC_POSY;
	DFBCHECK (pIterm->layer->CreateWindow(pIterm->layer, &desc, &pIterm->window));

	if (InitTerm () < 0)
		exit (TERM_ERROR_EXIT);

	if (InitVt () < 0)
		exit (VT_ERROR_EXIT);

	pthread_create (&thread, NULL, update_thread, (void*)pIterm);

	DFBCHECK (pIterm->window->GrabKeyboard(pIterm->window));
	DFBCHECK (pIterm->window->GrabPointer(pIterm->window));
	DFBCHECK (pIterm->window->CreateEventBuffer(pIterm->window, &pIterm->buffer));

	signal (SIGHUP, exitDFbiterm);
	signal (SIGTERM, exitDFbiterm);

	while (1) {
		DFBWindowEvent evt;

		pIterm->buffer->WaitForEvent(pIterm->buffer);
		while (pIterm->buffer->GetEvent(pIterm->buffer, DFB_EVENT(&evt)) == DFB_OK) {
			switch (evt.type) {
			case DWET_BUTTONDOWN:
			case DWET_BUTTONUP:
			case DWET_WHEEL:
				term_handle_button(&evt);
				break;
			case DWET_KEYUP:
			case DWET_KEYDOWN:
				term_handle_key(&evt);
				break;
			}
		}
	}
	return 0;
}

static void *
update_thread(void *arg)
{
	int fd = pIterm->fd;
	int ret;
	fd_set rfds;
	struct timeval timeout;

	while (1) {
		FD_ZERO (&rfds);
		FD_SET (fd, &rfds);
		timeout.tv_sec = 0;
		timeout.tv_usec = 100000;

		ret = select (fd + 1, &rfds, NULL, NULL, &timeout);
		if (ret == 0 || (ret < 0 && errno == EINTR)) continue;

		if (ret < 0) perror ("select");
		if (FD_ISSET (fd, &rfds)) {
			if (ret > 0) {
				VTCore_dispatch(pIterm->vtcore_ptr);
			}
		}
	}
	return NULL;
}
